(** OCamlFind tools *)

val debug : bool ref
(** Setting this [true] enables some debug printing to stderr *)

module Package : sig

  type t = { 
    name : string;
    dir  : string;
    defs : (string * string) list
  } [@@deriving conv{ocaml}]
  (** Simplified version of findlib's [Fl_package_base.package] *)

  val name : t -> string
  (** Get package name of [t] *)
  
  val find_var : string -> t -> string option
  (** Find a variable binding stored in [defs] field of [t] *)

  val version : t -> string option
  (** The version of the package like [Some "113.00.00"] *)

  val is_distributed_with_ocaml : t -> bool
  (** Check the version is "[distributed with Ocaml]" or something similar.

      Note that CamlP4 is not longer distributed with OCaml but it is still
      reported as it were.
  *)

  val is_base : t -> bool
  (** [is_base p], the package [p] is installed with OCaml compiler *)

  val top_name : t -> string
  (** returns the top package name: i.e. ["camlimages"] for camlimages.core *)

  val is_top : t -> bool
  (** [true] if the package is a top one. ex. camlimages.core is not top. *)

  val requires : t -> string list option
  (** Required packages for [t]. 
      [None] means the "requires" field itself is not found *)

  val requires_transitive : t list -> t -> t list
  (** [requires_transitive all t] finds the transitive closure of the required
      packages of [t]. [all] is the list of installed packages. *)

  val includes : t list -> t -> string list  
  (** [includes all t] returns include directories for the package [t].
      [all] is the list of installed packages *)
    
end

module Group : sig
  type t = {
    name : string; (** top_name *)
    packages : Package.t list
  } [@@deriving conv{ocaml}]
  (** Type for package groups, sharing the same top package name.
      For example, packages re, re.emacs, re.glob, re.pcre, re.perl, re.posix
      and re.str are grouped in this data type with [name = "re"]
  *)

  val top : t -> Package.t
  (** Get the top package *)

  val dir : t -> string
  (** Get the directory of the top package *)
end

val group : Package.t list -> Group.t list
(** Group packages by their top_names *)
    
module Analyzed : sig

  type t = {
    package           : Package.t;
    archives          : string list; (** archives for linking *)
    compilation_units : string list; (** compilation units accessible *)
    cmis              : string list; (** cmi files accessible *)
  } [@@deriving conv{ocaml}]
  (** [Package.t] with linking information *)

  val name : t -> string
  (** Return the name of the package *)
    
  val is_base : t -> bool
  (** Return [true] if it is a top package *)
end

module Analyzed_group : sig
  type t = {
    name     : string;
    packages       : Analyzed.t list;
    not_linked_cmis : string list; (** Not linked but accessible cmi's, such as cmis of
                                       mli only modules, ex. parsetree.cmi of compiler-libs
                                   *)
  } [@@deriving conv{ocaml}]
  (** Group of analyzed packages [Analyzed.t] which share the same top package
      name. 
  *)
      
  val dir : t -> string
  (** Get the directory of the top package *)
end

val analyze : Package.t -> Analyzed.t
(** Analyze a package *)

val analyze_group : stdlib_dir:string -> Group.t -> Analyzed_group.t
(** Analyze a group of packages, [stdlib_dir] is the location of 
    the standard library directory of ocaml system *)

val group_and_analyze : stdlib_dir:string -> Package.t list -> Analyzed_group.t list
(** [group] then [analyze_group] *)
  
val get : unit -> string * Package.t list
(** Get the stdlib dir and packages *)

val choose_best_package_name : string list -> string
(** Get the best (i.e. shortest) package name of many *)
    
val test : unit -> unit
(** Run some tests *)
